SuperStrict

Import "Octree.bmx"
Import "MatrixVector.bmx"
Include "Entity.bmx"

Const GWIDTH:Int  = 800
Const GHEIGHT:Int = 600

Const RAY_X_SCALE:Float = 1.0/GWIDTH
Const RAY_Y_SCALE:Float = 1.0/GHEIGHT

Const SAMPLE_SPACING:Float = 8

InitGL()

Local Tree:TVoxelOctree

'Set to one to convert volumetric data to octree
If 0 Then
	Local VoxelData:Byte[,,] = LoadVoxelData()
	Local Normals:Int[512, 512, 512]
	
	ExtractNormals(VoxelData, Normals)
	HighlightTissue(VoxelData, 115)
	
	Tree = New TVoxelOctree.Create(VoxelData, Normals)
	
	Tree.Save("StanfordBunny.oct")
Else
	Tree = New TVoxelOctree.Load("StanfordBunny.oct")
EndIf

Local Timer:TTimer = CreateTimer(60)

Local Scale:Float = 1.0, Yaw:Float

Local BufferPixmap:TPixmap = CreatePixmap(GWIDTH, GHEIGHT, PF_RGBA8888)

Local Volume:TEntity = New TEntity
Local LightDir:TVector4 = Vec3(1.0, -1.0, 1.0).Normalize()

Local DepthBuffer:Float[GWIDTH/SAMPLE_SPACING + 1, GHEIGHT/SAMPLE_SPACING + 1]
?Threaded
	Global Threads:TThread[(GWIDTH/SAMPLE_SPACING)*(GHEIGHT/SAMPLE_SPACING)]
?

While Not (KeyHit(KEY_ESCAPE) Or AppTerminate())
	Scale :+ MouseZSpeed()*0.05
	
	Volume.Scale(Scale, Scale, Scale)
	Volume.Locate(0.0, 0.0, 2.0)
	Volume.Turn(-MouseYSpeed(), MouseXSpeed(), 0.0)
	
	Local Time:Int = MilliSecs()
	
	If KeyDown(KEY_SPACE) Then
		glClear(GL_COLOR_BUFFER_BIT)
		glLoadIdentity()
		glScalef(1.0, 1.0, -1.0)
		glMultMatrixf(Volume.GetTransform().A)
		Tree.Render(-0.5, -0.5, -0.5)
	Else
		RenderScreen(Tree, Volume, LightDir, BufferPixmap, DepthBuffer)
	EndIf
	GLDrawText("Press spacebar for octree wireframe", 0, 0)
	
	Print MilliSecs() - Time
	
	Flip 0
	WaitTimer Timer
Wend
End

Function RenderScreen(Tree:TVoxelOctree, Volume:TEntity, LightDir:TVector4, BufferPixmap:TPixmap, DepthBuffer:Float[,])
	Const XScale:Float = 1.0/GWIDTH
	Const YScale:Float = 1.0/GHEIGHT
	
	Local Position :TVector4 = Vec3(0.0, 0.0, 0.0)
	Local Direction:TVector4 = Vec3(0.0, 0.0, 0.0)
	Local Normal   :TVector4 = Vec3(0.0, 0.0, 0.0)
	
	Local Buffer:Int Ptr = Int Ptr BufferPixmap.PixelPtr(0, 0)
	
	TEntity.TFormPoint (Position,  Null, Volume)
	For Local SY:Int = 0 To GHEIGHT/SAMPLE_SPACING
		For Local SX:Int = 0 To GWIDTH/SAMPLE_SPACING
			Direction.Set(SX*SAMPLE_SPACING*RAY_X_SCALE - 0.5, 0.5 - SY*SAMPLE_SPACING*RAY_Y_SCALE, 1.0)
			TEntity.TFormVector(Direction, Null, Volume)
			Direction.Normalize()
			
			Local IntNormal:Int
			DepthBuffer[SX, SY] = Tree.Raymarch(Position.X, Position.Y, Position.Z, Direction.X, Direction.Y, Direction.Z, 0.0, 0.0, IntNormal)
		Next
	Next
	
	Local N:Float[3], LocalLight:TVector4 = TFormVector(LightDir.Copy(), Null, Volume).VecMult(Volume.Scaling), ThreadIndex:Int
	For Local SY:Int = 0 Until GHEIGHT/SAMPLE_SPACING
		For Local SX:Int = 0 Until GWIDTH/SAMPLE_SPACING
			?Threaded
				Threads[ThreadIndex] = TThread.Create(RenderBatchThreaded, TRenderBatchInfo.Create(BufferPixmap, Volume, Tree, DepthBuffer, LocalLight, SX, SY))
				ThreadIndex :+ 1
			?Not Threaded
				RenderBatch(BufferPixmap, Volume, Tree, DepthBuffer, LocalLight, SX, SY)
			?
		Next
	Next
	
	?Threaded
		For Local T:TThread = EachIn Threads
			T.Wait()
		Next
	?
	
	GLDrawPixmap BufferPixmap, 0, 0
End Function

Type TRenderBatchInfo
	Field Target:TPixmap
	Field Volume:TEntity
	Field Tree:TVoxelOctree
	Field DepthBuffer:Float[,]
	Field LocalLight:TVector4
	Field SX:Int
	Field SY:Int
	
	Function Create:TRenderBatchInfo(Target:TPixmap, Volume:TEntity, Tree:TVoxelOctree, DepthBuffer:Float[,], LocalLight:TVector4, SX:Int, SY:Int)
		Local Info:TRenderBatchInfo = New TRenderBatchInfo
		
		Info.Target = Target
		Info.Volume = Volume
		Info.Tree = Tree
		Info.DepthBuffer = DepthBuffer
		Info.LocalLight = LocalLight
		Info.SX = SX
		Info.SY = SY
		
		Return Info
	End Function
End Type

Function RenderBatchThreaded:Object(Data:Object)
	Local Info:TRenderBatchInfo = TRenderBatchInfo(Data)
	
	RenderBatch(Info.Target, Info.Volume, Info.Tree, Info.DepthBuffer, Info.LocalLight, Info.SX, Info.SY)
	
	Return Null
End Function

Function RenderBatch(Target:TPixmap, Volume:TEntity, Tree:TVoxelOctree, DepthBuffer:Float[,], LocalLight:TVector4, SX:Int, SY:Int)
	Local Depth:Float = Min(DepthBuffer[SX, SY], Min(DepthBuffer[SX + 1, SY], Min(DepthBuffer[SX, SY + 1], DepthBuffer[SX + 1, SY + 1])))
	
	If Depth <> Tree.MISS Then
		Local N:Float[5]
		
		Local Position :TVector4 = Vec3(0.0, 0.0, 0.0)
		Local Direction:TVector4 = Vec3(0.0, 0.0, 0.0)
		Local Normal   :TVector4 = Vec3(0.0, 0.0, 0.0)
		
		For Local Y:Int = SY*SAMPLE_SPACING Until Min((SY + 1)*SAMPLE_SPACING, GHEIGHT)
			For Local X:Int = SX*SAMPLE_SPACING Until Min((SX + 1)*SAMPLE_SPACING, GWIDTH)
				Direction.Set(X*RAY_X_SCALE - 0.5, 0.5 - Y*RAY_Y_SCALE, 1.0)
				TEntity.TFormVector(Direction, Null, Volume)
				Direction.Normalize()
				
				Position .Set(0.0, 0.0, 0.0)
				TEntity.TFormPoint (Position,  Null, Volume)
				Position.Plus(Direction.Copy().Mult(Depth))
				
				Local IntNormal:Int
				Local T:Float = Tree.Raymarch(Position.X, Position.Y, Position.Z, Direction.X, Direction.Y, Direction.Z, 0.0, 0.0, IntNormal)
				
				DecompressNormal(IntNormal, N)
				
				If T = Tree.MISS Then
					Target.WritePixel(X, Y, 0)
				Else
					Position.Plus(Direction.Copy().Mult(T - 2.0/512.0))
					
					'Uncomment for shadows and stuffs
					'T = Tree.Raymarch(Position.X, Position.Y, Position.Z, -LocalLight.X, -LocalLight.Y, -LocalLight.Z, 0.0, 0.0, IntNormal)
					
					Normal.Set(N[0], N[1], N[2])
					
					Local Shade:Float = Max(Normal.Dot(LocalLight), 0.0)*0.8
					
					Shade :+ (Max(-LocalLight.Copy().Minus(Normal.Mult(2.0*Normal.Dot(LocalLight))).Dot(Direction), 0.0)^10.0)*0.2
					
					Shade = Min(Shade, 1.0)
					
					'If T <> Tree.MISS Then Shade :* 0.7'And T > 2.0/512.0 Then Shade :* 0.7
					
					Target.WritePixel(X, Y, Int(Shade*255.0)*$010101)
				EndIf
			Next
		Next
	Else
		For Local Y:Int = SY*SAMPLE_SPACING Until Min((SY + 1)*SAMPLE_SPACING, GHEIGHT)
			For Local X:Int = SX*SAMPLE_SPACING Until Min((SX + 1)*SAMPLE_SPACING, GWIDTH)
				Target.WritePixel(X, Y, 0)
			Next
		Next
	EndIf
End Function

Function InitGL()
	GLGraphics GWIDTH, GHEIGHT
	
	glewInit()
	
	glMatrixMode(GL_PROJECTION)
	glLoadIdentity()
	
	gluPerspective(60, GHEIGHT/Float(GWIDTH), 0.1, 20.0)
	
	'glTranslatef(0.0, 0.0, -5.0)
	
	glMatrixMode(GL_MODELVIEW)
	glLoadIdentity()
	
	'glEnable(GL_DEPTH_TEST)
End Function

'Extract isosurface from voxeldata (simple thresholding)
Function HighlightTissue(Data:Byte[,,], Threshold:Int)
	For Local X:Int = 0 Until 511
		For Local Y:Int = 0 Until 511
			For Local Z:Int = 0 Until 511
				Local B1:Int = (Data[Z, Y, X] > Threshold)
				Local B2:Int = (Data[Z, Y + 1, X] > Threshold)
				Local B3:Int = (Data[Z, Y, X + 1] > Threshold)
				Local B4:Int = (Data[Z + 1, Y, X] > Threshold)
				
				If B1 = B2 And B1 = B3 And B1 = B4 Then Data[Z, Y, X] = 0 Else Data[Z, Y, X] = 255
			Next
		Next
	Next
End Function

'Approximate gradient from voxel data
Function ExtractNormals(Data:Byte[,,], Dest:Int[,,])
	Local N:Float[5]
	
	Local UScale:Int = (1 Shl 16) - 1
	Local VScale:Int = (1 Shl 15) - 1
	
	For Local X:Int = 1 Until 511
		For Local Y:Int = 1 Until 511
			For Local Z:Int = 1 Until 511
				Local D1:Float = (Data[X + 1, Y, Z] - Data[X - 1, Y, Z])*0.00390625
				Local D2:Float = (Data[X, Y + 1, Z] - Data[X, Y - 1, Z])*0.00390625
				Local D3:Float = (Data[X, Y, Z + 1] - Data[X, Y, Z - 1])*0.00390625
				
				Local InvLength:Float = 1.0/Sqr(D1*D1 + D2*D2 + D3*D3)
				
				N[0] = D1*InvLength
				N[1] = D2*InvLength
				N[2] = D3*InvLength
				N[3] = N[0]
				N[4] = N[1]
				
				Dest[X, Y, Z] = CompressNormal(N)
			Next
		Next
		
		Print (100.0*X/512.0) + "% finished!"
	Next
End Function

'Compress normal into 32 bits
Function CompressNormal:Int(N:Float[])
	Local UScale:Int = (1 Shl 15) - 1
	Local VScale:Int = (1 Shl 14) - 1
	
	Local DominantDirection:Float = -1.0, Face:Int
	
	For Local I:Int = 0 Until 3
		If Abs(N[I]) > DominantDirection Then
			Face = I
			DominantDirection = Abs(N[I])
		EndIf
	Next
	
	Local Sign:Int = (N[Face] < 0.0)
	
	N[Face + 1] :/ Abs(N[Face])
	N[Face + 2] :/ Abs(N[Face])
	
	Local U:Int = Int((N[Face + 1]*0.5 + 0.5)*UScale)
	Local V:Int = Int((N[Face + 2]*0.5 + 0.5)*VScale)
	
	Return ((Sign Shl 31) | (Face Shl 29) | (U Shl 14) | V)
End Function

'Uncompress normal back to 3 floats
Function DecompressNormal(Normal:Int, Dest:Float[])
	Local Sign:Int = (Normal & $80000000) Shr 31
	Local Face:Int = (Normal & $60000000) Shr 29
	Local U   :Int = (Normal & $1FFFC000) Shr 14
	Local V   :Int = (Normal & $00003FFF)
	
	Dest[Face] = 1.0 - Sign*2.0
	Dest[(Face + 1) Mod 3] = U*3.0519e-5*2.0 - 1.0
	Dest[(Face + 2) Mod 3] = V*6.1037e-5*2.0 - 1.0
	
	Local InvLength:Float = 1.0/Sqr(Dest[0]*Dest[0] + Dest[1]*Dest[1] + Dest[2]*Dest[2])
	
	Dest[0] :* InvLength
	Dest[1] :* InvLength
	Dest[2] :* InvLength
End Function

Function LoadVoxelData:Byte[,,]()
	Local Data:Byte[512, 512, 512]
	
	Local Instream:TStream = ReadFile("StanfordBunny.voxel")
	Instream.ReadBytes(Varptr Data[0, 0, 0], 512*512*512)
	Instream.Close()
	
	Return Data
End Function